import logging
from collections import deque
from time import time

from orbitant.transport import TransportSlave


class BpmFinder(TransportSlave):

    def __init__(self, port, min_samples, max_samples):
        super(BpmFinder, self).__init__(port)
        self.min_samples = min_samples
        self.queue = deque(maxlen=max_samples)
        self.last_bpm = None

    @property
    def current_bpm(self):
        if len(self.queue) >= self.min_samples:
            return (
                60.
                /
                (
                    (self.queue[-1] - self.queue[0]) /
                    len(self.queue) *
                    4
                )
            )

    def on_16th(self):
        self.queue.append(time())
        bpm = int(self.current_bpm or 0)
        if bpm and bpm != self.last_bpm:
            logging.info('bpm=%r', bpm)
            self.last_bpm = bpm



import logging

import begin

from orbitant import backend, generator


@begin.start
@begin.logging
def main(bpm: 'Beats per Minute' = 120,
         tpb: 'Ticks per Beat' = 24,
         name: 'MIDI port name' = 'Orbitant Generator',
         ):
    bpm = int(bpm)
    tpb = int(tpb)
    port = backend.open_output(name=name, virtual=True)
    gen = generator.Generator(port=port, bpm=bpm, tpb=tpb)
    logging.info('Starting clock; press ^C to exit')
    gen.start()



import mido


CLOCK = mido.Message('clock')
START = mido.Message('start')
STOP = mido.Message('stop')


TICKS_32ND = 96 // 32
TICKS_SEXTUPLET = 96 // 24
TICKS_16TH = 96 // 16
TICKS_TRIPLET = 96 // 12
TICKS_8TH = 96 // 8
TICKS_QUARTER = 96 // 4
TICKS_HALF = 96 // 2
TICKS_WHOLE = 96 // 1


def beat_properties(ticks):

    @property
    def beat(self):
        return self.tick // ticks

    @property
    def beat_tick(self):
        return self.tick % ticks

    return beat, beat_tick


class TransportSlave(object):
    """Listens and reacts to MIDI beat clock and transport messages.

    Subclass this and override the handlers to implement desired functionality.

    Supports Tick, Song Position Pointer (0, 0), Start, Stop, and Continue.
    """

    def __init__(self, port):
        self.port = port
        self._started = False
        self._tick = 0

    @property
    def started(self):
        return self._started

    @property
    def tick(self):
        return self._tick

    midi_beat, midi_beat_tick = beat_properties(6)

    beat_32nd, beat_32nd_tick = beat_properties(TICKS_32ND)
    beat_sextuplet, beat_sextuplet_tick = beat_properties(TICKS_SEXTUPLET)
    beat_16th, beat_16th_tick = beat_properties(TICKS_16TH)
    beat_triplet, beat_triplet_tick = beat_properties(TICKS_TRIPLET)
    beat_8th, beat_8th_tick = beat_properties(TICKS_8TH)
    beat_quarter, beat_quarter_tick = beat_properties(TICKS_QUARTER)
    beat_half, beat_half_tick = beat_properties(TICKS_HALF)
    beat_whole, beat_whole_tick = beat_properties(TICKS_WHOLE)

    @property
    def midi_beat(self):
        return self.tick // 6

    @property
    def midi_beat_tick(self):
        return self.tick % 6

    def run(self):
        for message in self.port:
            if message.type == 'clock':
                self._increment_tick()
            elif message.type == 'start':
                self._started = True
                self.on_start()
            elif message.type == 'stop':
                self._started = False
                self.on_stop()
            elif message.type == 'continue':
                self._started = True
                self.on_continue()
            elif message.type == 'songpos' and message.pos == 0:
                self._tick = 0
                self.on_rewind()

    def _increment_tick(self):
        self.on_tick()
        t = self.tick
        if t % TICKS_32ND == 0:
            self.on_32nd()
        if t % TICKS_SEXTUPLET == 0:
            self.on_sextuplet()
        if t % TICKS_16TH == 0:
            self.on_16th()
        if t % TICKS_TRIPLET == 0:
            self.on_triplet()
        if t % TICKS_8TH == 0:
            self.on_8th()
        if t % TICKS_QUARTER == 0:
            self.on_quarter()
        if t % TICKS_HALF == 0:
            self.on_half()
        if t % TICKS_WHOLE == 0:
            self.on_whole()
        self._tick += 1

    # Override these in subclasses:
    def on_start(self): pass

    def on_stop(self): pass

    def on_continue(self): pass

    def on_rewind(self): pass

    def on_tick(self): pass

    def on_midi_beat(self): pass

    def on_32nd(self): pass

    def on_sextuplet(self): pass

    def on_16th(self): pass

    def on_triplet(self): pass

    def on_8th(self): pass

    def on_quarter(self): pass

    def on_half(self): pass

    def on_whole(self): pass



import logging
from time import sleep, time

import mido


CLOCK = mido.Message('clock')


class Generator(object):

    def __init__(self, port, bpm, tpb):
        self.port = port
        self.bpm = bpm
        self.tpb = tpb
        self.ticks = 0

    def start(self):
        initial_time = time()
        tick_time = 60. / self.bpm / self.tpb
        wait_time = tick_time / 1.5
        sleep_time = wait_time / 2.
        while True:
            next_time = initial_time + (tick_time * self.ticks)
            while time() + wait_time < next_time:
                sleep(sleep_time)
            while time() < next_time:
                pass
            self.tick()
            self.ticks += 1

    @property
    def beat(self):
        return self.ticks // self.tpb

    @property
    def beat_tick(self):
        return self.ticks % self.tpb

    def tick(self):
        self.port.send(CLOCK)
        logging.debug('beat=%i tick=%i', self.beat, self.beat_tick)



import logging

import begin

from orbitant import backend, bpmfinder


@begin.start
@begin.logging
def main(min_samples: 'Timing samples to accumulate before calculating' = 16,
         max_samples: 'Maximum timing samples to use when calculating' = 128,
         name: 'MIDI port name' = 'Orbitant Generator'):
    logging.info('Starting monitor; press ^C to exit')
    with backend.open_input(name=name) as port:
        bpm_finder = bpmfinder.BpmFinder(
            port=port,
            min_samples=int(min_samples),
            max_samples=int(max_samples),
        )
        bpm_finder.run()



"""MIDI clock utilities for Python."""

import io
import os
import re
import sys
from setuptools import find_packages, setup

SETUP_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__)))
sys.path.append(SETUP_DIR)
import orbitant  # NOQA isort:skip

dependencies = [
    'begins',
    'mido',
    'python-rtmidi',
]


def read(*names, **kwargs):
    return io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


setup(
    name='orbitant',
    version=orbitant.__version__,
    url='https://github.com/metrasynth/orbitant',
    license='MIT',
    author='Matthew Scott',
    author_email='matt@11craft.com',
    description=__doc__,
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).
            sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    packages=find_packages(exclude=['docs', 'tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    entry_points={},
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
